 

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteOrder;
import java.util.ArrayList;

public class RandomAccessFileSP extends RandomAccessFile {
	public static final boolean RANDOMACCESSBUFFER = false;
	public static final boolean SETBUFFER = true;
	public static final ByteOrder READER_WRITER_BYTE_ORDER = ByteOrder.BIG_ENDIAN;
	
	public ArrayList<Shift> shifts;
	
	boolean bufferType = SETBUFFER;
	public byte[] fileBuffer;
	int bufferSize = -1;
	int attemptBufferSize = -1;
	ByteOrder byteOrder = ByteOrder.nativeOrder();
	
	long bufferOffset = 0;
	int offsetInBuffer = 0;
	long fileOffset = 0;
	
	public RandomAccessFileSP(File file, String fileAccess) throws FileNotFoundException{
		super(file, fileAccess);
		fileBuffer = new byte[0];
		this.shifts = new ArrayList<Shift>();
		
	}
	
	public RandomAccessFileSP(File file, String fileAccess, int bufferSize) throws FileNotFoundException{
		super(file, fileAccess);
		bufferType = RANDOMACCESSBUFFER;
		this.attemptBufferSize = bufferSize;
		this.shifts = new ArrayList<Shift>();
	}
	
	public void bufferSeek(int offset){
		this.offsetInBuffer = offset;
	}
	
	public void setByteOrder(ByteOrder byteOrder){
		this.byteOrder = byteOrder;
	}
	
	public void seek(long offset) throws IOException{
		super.seek(offset);
		this.fileOffset = offset;
	}
	
	public byte getByte() throws IOException{
		byte[]data = getBytes(this.fileOffset, 1);
		return data[0];
	}
	
	public short getUByte() throws IOException{
		byte[]data = getBytes(this.fileOffset, 1);
		return data[0];
	}
	
	public short getShort() throws IOException{
		byte[] data = correctOrderByteArray(getBytes(this.fileOffset, 2));
		return (short)(((data[0] << 8 ) & 0x0000FF00) |
				       ( data[1]        & 0x000000FF));
	}
	
	public int getUShort() throws IOException{
		byte[] data = correctOrderByteArray(getBytes(this.fileOffset, 2));
		return (int)(((data[0] << 8 ) & 0x0000FF00) |
				     ( data[1]        & 0x000000FF));
	}
	
	public int getInt() throws IOException{
		byte[] data = correctOrderByteArray(getBytes(this.fileOffset, 4));
		return (((data[0] << 24) & 0xFF000000) |
				((data[1] << 16) & 0x00FF0000) |
				((data[2] << 8 ) & 0x0000FF00) |
				( data[3]        & 0x000000FF));
	}
	
	public long getUInt() throws IOException{
		byte[] data = correctOrderByteArray(getBytes(this.fileOffset, 4));
		return (long)(((data[0] << 24) & 0xFF000000L) |
				      ((data[1] << 16) & 0x00FF0000L) |
				      ((data[2] << 8 ) & 0x0000FF00L) |
				      ( data[3]        & 0x000000FFL));
	}
	
	public long getLong() throws IOException{
		byte[] data = correctOrderByteArray(getBytes(this.fileOffset, 8));
		return (long)(((data[0] << 56) & 0xFF00000000000000L) |
			      ((data[1] << 48) & 0x00FF000000000000L) |
			      ((data[2] << 40) & 0x0000FF0000000000L) |
			      ((data[3] << 32) & 0x000000FF00000000L) |
			      ((data[4] << 24) & 0x00000000FF000000L) |
			      ((data[5] << 16) & 0x0000000000FF0000L) |
			  	  ((data[6] << 8 ) & 0x000000000000FF00L) |
			  	   (data[7]        & 0x00000000000000FFL));
	}
	
	public float getFloat() throws IOException{
		int i = getInt();
		return Float.intBitsToFloat(i);
	}
	
	public String getString() throws IOException{
		String string = "";
		byte[] data = getBytes(this.fileOffset, 1);
		while(data[0] != 0){
			string += (char)data[0];
			data = getBytes(this.fileOffset, 1);
		}
		return string;
	}
	
	public String getString(int length, boolean reverse) throws IOException{
		String string ="";
		byte[] data = getBytes(this.fileOffset, length);
		for(int i = 0; i < length; i++){
			if(!reverse){
				if(data[i] == 0){
					break;
				}
				string += (char)data[i];
			}else{
				string = (char)data[i] + string;
			}
		}
		return string;
	}
	
	public void putByte(byte b) throws IOException{
		writeBytes(this.fileOffset, new byte[]{b});
	}
	
	public void putUByte(short b) throws IOException{
		writeBytes(this.fileOffset, new byte[]{(byte)b});
	}
	
	public void putShort(short s) throws IOException{
		byte[] data = new byte[2];
		data[0] = (byte)((s >> 8) & 0xFF);
		data[1] = (byte)(s & 0xFF);
		writeBytes(this.fileOffset, correctOrderByteArray(data));
	}
	
	public void putUShort(short s) throws IOException{
		byte[] data = new byte[2];
		data[0] = (byte)((s >> 8) & 0xFF);
		data[1] = (byte)(s & 0xFF);
		writeBytes(this.fileOffset, correctOrderByteArray(data));
	}
	
	public void putInt(int i) throws IOException{
		byte[] data = new byte[4];
		data[0] = (byte)((i >> 24) & 0xFF);
		data[1] = (byte)((i >> 16) & 0xFF);
		data[2] = (byte)((i >> 8) & 0xFF);
		data[3] = (byte)(i & 0xFF);
		writeBytes(this.fileOffset,  correctOrderByteArray(data));
	}
	
	public void putUInt(long i) throws IOException{
		byte[] data = new byte[4];
		data[0] = (byte)((i >> 24) & 0xFF);
		data[1] = (byte)((i >> 16) & 0xFF);
		data[2] = (byte)((i >> 8) & 0xFF);
		data[3] = (byte)(i & 0xFF);
		writeBytes(this.fileOffset,  correctOrderByteArray(data));
	}
	
	public void putLong(long l) throws IOException{
		byte[] data = new byte[8];
		data[0] = (byte)((l >> 56) & 0xFF);
		data[1] = (byte)((l >> 48) & 0xFF);
		data[2] = (byte)((l >> 40) & 0xFF);
		data[3] = (byte)((l >> 32) & 0xFF);
		data[4] = (byte)((l >> 24) & 0xFF);
		data[5] = (byte)((l >> 16) & 0xFF);
		data[6] = (byte)((l >> 8) & 0xFF);
		data[7] = (byte)(l & 0xFF);
		writeBytes(this.fileOffset,  correctOrderByteArray(data));
	}
	
	public void putFloat(float f) throws IOException{
		int i = Float.floatToIntBits(f);
		putInt(i);
	}
	
	public void putString(String string) throws IOException{
		byte[] data = new byte[string.length() + 1];
		for(int i = 0; i < string.length(); i++){
			data[i] = (byte)string.charAt(i);
		}
		writeBytes(this.fileOffset, data);
	}
	
	public void putString(String string, int length, boolean reverse) throws IOException{
		byte[] data = new byte[length];
		for(int i = 0; i < string.length(); i++){
			if(!reverse){
				data[i] = (byte)string.charAt(i);
			}else{
				data[data.length - 1 - i] = (byte)string.charAt(i);
			}
		}
		writeBytes(this.fileOffset, data);
	}
	
	public byte getByte(long offset) throws IOException{
		byte[]data = getBytes(offset, 1);
		return data[0];
	}
	
	public short getUByte(long offset) throws IOException{
		byte[]data = getBytes(offset, 1);
		return data[0];
	}
	
	public short getShort(long offset) throws IOException{
		byte[] data = correctOrderByteArray(getBytes(offset, 2));
		return (short)(((data[0] << 8 ) & 0x0000FF00) |
				       ( data[1]        & 0x000000FF));
	}
	
	public int getUShort(long offset) throws IOException{
		byte[] data = correctOrderByteArray(getBytes(offset, 2));
		return (int)(((data[0] << 8 ) & 0x0000FF00) |
				     ( data[1]        & 0x000000FF));
	}
	
	public int getInt(long offset) throws IOException{
		byte[] data = correctOrderByteArray(getBytes(offset, 4));
		return (((data[0] << 24) & 0xFF000000) |
				((data[1] << 16) & 0x00FF0000) |
				((data[2] << 8 ) & 0x0000FF00) |
				( data[3]        & 0x000000FF));
	}
	
	public long getUInt(long offset) throws IOException{
		byte[] data = correctOrderByteArray(getBytes(offset, 4));
		return (long)(((data[0] << 24) & 0xFF000000L) |
				 	  ((data[1] << 16) & 0x00FF0000L) |
				      ((data[2] << 8 ) & 0x0000FF00L) |
				      ( data[3]        & 0x000000FFL));
	}
	
	public long getLong(long offset) throws IOException{
		byte[] data = correctOrderByteArray(getBytes(offset, 8));
		return (long)(((data[0] << 56) & 0xFF00000000000000L) |
			      ((data[1] << 48) & 0x00FF000000000000L) |
			      ((data[2] << 40) & 0x0000FF0000000000L) |
			      ((data[3] << 32) & 0x000000FF00000000L) |
			      ((data[4] << 24) & 0x00000000FF000000L) |
			      ((data[5] << 16) & 0x0000000000FF0000L) |
			  	  ((data[6] << 8 ) & 0x000000000000FF00L) |
			  	   (data[7]        & 0x00000000000000FFL));
	}
	
	public float getFloat(long offset) throws IOException{
		int i = getInt(offset);
		return Float.intBitsToFloat(i);
	}
	
	public String getString(long offset) throws IOException{
		String string = "";
		byte[] data = getBytes(offset, 1);
		while(data[0] != 0){
			string += (char)data[0];
			data = getBytes(this.fileOffset, 1);
		}
		return string;
	}
	
	public String getString(long offset, int length, boolean reverse) throws IOException{
		String string ="";
		byte[] data = getBytes(offset, length);
		for(int i = 0; i < length; i++){
			if(!reverse){
				if(data[i] == 0){
					break;
				}
				string += (char)data[i];
			}else{
				string = (char)data[i] + string;
			}
		}
		return string;
	}
	
	public void putByte(long offset, byte b) throws IOException{
		writeBytes(offset, new byte[]{b});
	}
	
	public void putUByte(long offset, short b) throws IOException{
		writeBytes(offset, new byte[]{(byte)b});
	}
	
	public void putShort(long offset, short s) throws IOException{
		byte[] data = new byte[2];
		data[0] = (byte)((s >> 8) & 0xFF);
		data[1] = (byte)(s & 0xFF);
		writeBytes(offset, correctOrderByteArray(data));
	}
	
	public void putUShort(long offset, short s) throws IOException{
		byte[] data = new byte[2];
		data[0] = (byte)((s >> 8) & 0xFF);
		data[1] = (byte)(s & 0xFF);
		writeBytes(offset, correctOrderByteArray(data));
	}
	
	public void putInt(long offset, int i) throws IOException{
		byte[] data = new byte[4];
		data[0] = (byte)((i >> 24) & 0xFF);
		data[1] = (byte)((i >> 16) & 0xFF);
		data[2] = (byte)((i >> 8) & 0xFF);
		data[3] = (byte)(i & 0xFF);
		writeBytes(offset,  correctOrderByteArray(data));
	}
	
	public void putUInt(long offset, long i) throws IOException{
		byte[] data = new byte[4];
		data[0] = (byte)((i >> 24) & 0xFF);
		data[1] = (byte)((i >> 16) & 0xFF);
		data[2] = (byte)((i >> 8) & 0xFF);
		data[3] = (byte)(i & 0xFF);
		writeBytes(offset,  correctOrderByteArray(data));
	}
	
	public void putLong(long offset, long l) throws IOException{
		byte[] data = new byte[8];
		data[0] = (byte)((l >> 56) & 0xFF);
		data[1] = (byte)((l >> 48) & 0xFF);
		data[2] = (byte)((l >> 40) & 0xFF);
		data[3] = (byte)((l >> 32) & 0xFF);
		data[4] = (byte)((l >> 24) & 0xFF);
		data[5] = (byte)((l >> 16) & 0xFF);
		data[6] = (byte)((l >> 8) & 0xFF);
		data[7] = (byte)(l & 0xFF);
		writeBytes(offset,  correctOrderByteArray(data));
	}
	
	public void putFloat(long offset, float f) throws IOException{
		int i = Float.floatToIntBits(f);
		putInt(offset, i);
	}
	
	public void putString(long offset, String string) throws IOException{
		byte[] data = new byte[string.length() + 1];
		for(int i = 0; i < string.length(); i++){
			data[i] = (byte)string.charAt(i);
		}
		writeBytes(offset, data);
	}
	
	public void putString(long offset, String string, int length, boolean reverse) throws IOException{
		byte[] data = new byte[length];
		for(int i = 0; i < string.length(); i++){
			if(!reverse){
				data[i] = (byte)string.charAt(i);
			}else{
				data[data.length - 1 - i] = (byte)string.charAt(i);
			}
		}
		writeBytes(offset, data);
	}
	
	public void bufferSkip(int size){
		this.offsetInBuffer += size;
	}
	
	public byte getBuffedByte() {
		byte[] data = readBuffedBytes(this.offsetInBuffer, 1);
		return data[0];
	}
	
	public short getBuffedUByte(){
		byte[] data = readBuffedBytes(this.offsetInBuffer, 1);
		return (short)data[0];
	}
	
	public short getBuffedShort() {
		byte[] data = correctOrderByteArray(readBuffedBytes(this.offsetInBuffer, 2));
		return (short)(((data[0] << 8 ) & 0x0000FF00) |
					   ( data[1]        & 0x000000FF));
	}
	
	public int getBuffedUShort() {
		byte[] data = correctOrderByteArray(readBuffedBytes(this.offsetInBuffer, 2));
		return (int)(((data[0] << 8 ) & 0x0000FF00) |
					 ( data[1]        & 0x000000FF));
	}
	
	public int getBuffedInt(){
		byte[] data = correctOrderByteArray(readBuffedBytes(this.offsetInBuffer, 4));
		return (int)(((data[0] << 24) & 0xFF000000) |
					 ((data[1] << 16) & 0x00FF0000) |
					 ((data[2] << 8 ) & 0x0000FF00) |
					 ( data[3]        & 0x000000FF));
	}
	
	public long getBuffedUInt() {
		byte[] data = correctOrderByteArray(readBuffedBytes(this.offsetInBuffer, 4));
		return (long)(((data[0] << 24) & 0xFF000000L) |
					 ((data[1] << 16) & 0x00FF0000L) |
					 ((data[2] << 8 ) & 0x0000FF00L) |
					 ( data[3]        & 0x000000FFL));
	}
	
	public long getBuffedLong(){
		byte[] data = correctOrderByteArray(readBuffedBytes(this.offsetInBuffer, 8));
		return (long)(((data[0] << 56) & 0xFF00000000000000L) |
					  ((data[1] << 48) & 0x00FF000000000000L) |
					  ((data[2] << 40) & 0x0000FF0000000000L) |
					  ((data[3] << 32) & 0x000000FF00000000L) |
					  ((data[4] << 24) & 0x00000000FF000000L) |
					  ((data[5] << 16) & 0x0000000000FF0000L) |
					  ((data[6] << 8 ) & 0x000000000000FF00L) |
					  	      (data[7] & 0x00000000000000FFL));
	}
	
	public float getBuffedFloat() {
		int i = getBuffedInt();
		return Float.intBitsToFloat(i);
	}
	
	public String getBuffedString() {
		String string = "";
		byte[] data = readBuffedBytes(this.offsetInBuffer, 1);
		while(data[0] != 0){
			string += (char)data[0];
			data = readBuffedBytes(this.offsetInBuffer, 1);
		}
		return string;
	}
	
	public String getBuffedString(int length, boolean reverse){
		String string ="";
		byte[] data = readBuffedBytes(this.offsetInBuffer, length);
		for(int i = 0; i < length; i++){
			if(!reverse){
				if(data[i] == 0){
					break;
				}
				string += (char)data[i];
			}else{
				string = (char)data[i] + string;
			}
		}
		return string;
	}
	
	public void putBuffedByte(byte b) {
		this.writeBuffedBytes(this.offsetInBuffer, new byte[]{b});
	}
	
	public void putBuffedUByte(short b) {
		this.writeBuffedBytes(this.offsetInBuffer, new byte[]{(byte)b});
	}
	
	public void putBuffedShort(short s) {
		byte[] data = new byte[2];
		data[0] = (byte)((s >> 8) & 0xFF);
		data[1] = (byte)(s & 0xFF);
		this.writeBuffedBytes(this.offsetInBuffer, correctOrderByteArray(data));
	}
	
	public void putBuffedUShort(short s) {
		byte[] data = new byte[2];
		data[0] = (byte)((s >> 8) & 0xFF);
		data[1] = (byte)(s & 0xFF);
		this.writeBuffedBytes(this.offsetInBuffer, correctOrderByteArray(data));
	}
	
	public void putBuffedInt(int i) {
		byte[] data = new byte[4];
		data[0] = (byte)((i >> 24) & 0xFF);
		data[1] = (byte)((i >> 16) & 0xFF);
		data[2] = (byte)((i >> 8) & 0xFF);
		data[3] = (byte)(i & 0xFF);
		this.writeBuffedBytes(this.offsetInBuffer,  correctOrderByteArray(data));
	}
	
	public void putBuffedUInt(long i) {
		byte[] data = new byte[4];
		data[0] = (byte)((i >> 24) & 0xFF);
		data[1] = (byte)((i >> 16) & 0xFF);
		data[2] = (byte)((i >> 8) & 0xFF);
		data[3] = (byte)(i & 0xFF);
		this.writeBuffedBytes(this.offsetInBuffer,  correctOrderByteArray(data));
	}
	
	public void putBuffedLong(long l) {
		byte[] data = new byte[8];
		data[0] = (byte)((l >> 56) & 0xFF);
		data[1] = (byte)((l >> 48) & 0xFF);
		data[2] = (byte)((l >> 40) & 0xFF);
		data[3] = (byte)((l >> 32) & 0xFF);
		data[4] = (byte)((l >> 24) & 0xFF);
		data[5] = (byte)((l >> 16) & 0xFF);
		data[6] = (byte)((l >> 8) & 0xFF);
		data[7] = (byte)(l & 0xFF);
		this.writeBuffedBytes(this.offsetInBuffer,  correctOrderByteArray(data));
	}
	
	public void putBuffedFloat(float f) {
		int i = Float.floatToIntBits(f);
		putBuffedInt(i);
	}
	
	public void putBuffedString(String string) {
		byte[] data = new byte[string.length() + 1];
		for(int i = 0; i < string.length(); i++){
			data[i] = (byte)string.charAt(i);
		}
		this.writeBuffedBytes(this.offsetInBuffer, data);
	}
	
	public void putBuffedString(String string, int length, boolean reverse) {
		byte[] data = new byte[length];
		for(int i = 0; i < string.length(); i++){
			if(!reverse){
				data[i] = (byte)string.charAt(i);
			}else{
				data[data.length - 1 - i] = (byte)string.charAt(i);
			}
		}
		this.writeBuffedBytes(this.offsetInBuffer, data);
	}
	
	public byte getBuffedByte(int offset) {
		byte[] data = readBuffedBytes(offset, 1);
		return data[0];
	}
	
	public short getBuffedUByte(int offset){
		byte[] data = readBuffedBytes(offset, 1);
		return (short)data[0];
	}
	
	public short getBuffedShort(int offset) {
		byte[] data = correctOrderByteArray(readBuffedBytes(offset, 2));
		return (short)(((data[0] << 8 ) & 0x0000FF00) |
				       ( data[1]        & 0x000000FF));
	}
	
	public int getBuffedUShort(int offset) {
		byte[] data = correctOrderByteArray(readBuffedBytes(offset, 2));
		return (int)(((data[0] << 8 ) & 0x0000FF00) |
				     ( data[1]        & 0x000000FF));
	}
	
	public int getBuffedInt(int offset){
		byte[] data = correctOrderByteArray(readBuffedBytes(offset, 4));
		return (((data[0] << 24) & 0xFF000000) |
				((data[1] << 16) & 0x00FF0000) |
				((data[2] << 8 ) & 0x0000FF00) |
				( data[3]        & 0x000000FF));
	}
	
	public long getBuffedUInt(int offset) {
		byte[] data = correctOrderByteArray(readBuffedBytes(offset, 4));
		return (long)(((data[0] << 24) & 0xFF000000L) |
				 ((data[1] << 16) & 0x00FF0000L) |
				 ((data[2] << 8 ) & 0x0000FF00L) |
				 ( data[3]        & 0x000000FFL));
	}
	
	public long getBuffedLong(int offset){
		byte[] data = correctOrderByteArray(readBuffedBytes(offset, 8));
		return (long)(((data[0] << 56) & 0xFF00000000000000L) |
				      ((data[1] << 48) & 0x00FF000000000000L) |
				      ((data[2] << 40) & 0x0000FF0000000000L) |
				      ((data[3] << 32) & 0x000000FF00000000L) |
				      ((data[4] << 24) & 0x00000000FF000000L) |
				      ((data[5] << 16) & 0x0000000000FF0000L) |
				  	  ((data[6] << 8 ) & 0x000000000000FF00L) |
				  	   (data[7]        & 0x00000000000000FFL));
	}
	
	public float getBuffedFloat(int offset) throws IOException{
		int i = getBuffedInt(offset);
		return Float.intBitsToFloat(i);
	}
	
	public String getBuffedString(int offset){
		String string = "";
		byte[] data = readBuffedBytes(offset, 1);
		while(data[0] != 0){
			string += (char)data[0];
			data = readBuffedBytes(this.offsetInBuffer, 1);
		}
		return string;
	}
	
	public String getBuffedString(int offset, int length, boolean reverse){
		String string ="";
		byte[] data = readBuffedBytes(offset, length);
		for(int i = 0; i < length; i++){
			if(!reverse){
				if(data[i] == 0){
					break;
				}
				string += (char)data[i];
			}else{
				string = (char)data[i] + string;
			}
		}
		return string;
	}
	
	public void putBuffedByte(int offset, byte b) {
		this.writeBuffedBytes(offset, new byte[]{b});
	}
	
	public void putBuffedUByte(int offset, short b) {
		this.writeBuffedBytes(offset, new byte[]{(byte)b});
	}
	
	public void putBuffedShort(int offset, short s) {
		byte[] data = new byte[2];
		data[0] = (byte)((s >> 8) & 0xFF);
		data[1] = (byte)(s & 0xFF);
		this.writeBuffedBytes(offset, correctOrderByteArray(data));
	}
	
	public void putBuffedUShort(int offset, short s) {
		byte[] data = new byte[2];
		data[0] = (byte)((s >> 8) & 0xFF);
		data[1] = (byte)(s & 0xFF);
		this.writeBuffedBytes(offset, correctOrderByteArray(data));
	}
	
	public void putBuffedInt(int offset, int i) {
		byte[] data = new byte[4];
		data[0] = (byte)((i >> 24) & 0xFF);
		data[1] = (byte)((i >> 16) & 0xFF);
		data[2] = (byte)((i >> 8) & 0xFF);
		data[3] = (byte)(i & 0xFF);
		this.writeBuffedBytes(offset,  correctOrderByteArray(data));
	}
	
	public void putBuffedUInt(int offset, long i) {
		byte[] data = new byte[4];
		data[0] = (byte)((i >> 24) & 0xFF);
		data[1] = (byte)((i >> 16) & 0xFF);
		data[2] = (byte)((i >> 8) & 0xFF);
		data[3] = (byte)(i & 0xFF);
		this.writeBuffedBytes(offset,  correctOrderByteArray(data));
	}
	
	public void putBuffedLong(int offset, long l) {
		byte[] data = new byte[8];
		data[0] = (byte)((l >> 56) & 0xFF);
		data[1] = (byte)((l >> 48) & 0xFF);
		data[2] = (byte)((l >> 40) & 0xFF);
		data[3] = (byte)((l >> 32) & 0xFF);
		data[4] = (byte)((l >> 24) & 0xFF);
		data[5] = (byte)((l >> 16) & 0xFF);
		data[6] = (byte)((l >> 8) & 0xFF);
		data[7] = (byte)(l & 0xFF);
		this.writeBuffedBytes(offset,  correctOrderByteArray(data));
	}
	
	public void putBuffedFloat(int offset, float f) {
		int i = Float.floatToIntBits(f);
		putBuffedInt(offset, i);
	}
	
	public void putBuffedString(int offset, String string) {
		byte[] data = new byte[string.length() + 1];
		for(int i = 0; i < string.length(); i++){
			data[i] = (byte)string.charAt(i);
		}
		this.writeBuffedBytes(offset, data);
	}
	
	public void putBuffedString(int offset, String string, int length, boolean reverse) {
		byte[] data = new byte[length];
		for(int i = 0; i < string.length(); i++){
			if(!reverse){
				data[i] = (byte)string.charAt(i);
			}else{
				data[data.length - 1 - i] = (byte)string.charAt(i);
			}
		}
		this.writeBuffedBytes(offset, data);
	}
	
	public byte[] correctOrderByteArray(byte[] old){
		if(this.byteOrder.equals(ByteOrder.LITTLE_ENDIAN)){
			byte[] data = new byte[old.length];
			int pos = 0;
			for(int i = old.length - 1; i > -1; i--){
				data[pos] = old[i];
				pos++;
			}
			return data;
		}
		return old;
	}
	
	public byte[] readBuffedBytes(int offset, int size){
		byte[] data = new byte[size];
		for(int i = 0; i < size; i++){
			data[i] = this.fileBuffer[offset + i];
		}
		this.offsetInBuffer = offset + size;
		return data;
	}
	
	public void writeBuffedBytes(int offset, byte[] data){
		for(int i = 0; i < data.length; i++){
			this.fileBuffer[offset + i] = data[i];
		}
		this.offsetInBuffer = offset + data.length;
	}
	
	public byte[] getBytes(long offset, int amount) throws IOException{
		if(this.bufferType == RANDOMACCESSBUFFER){
			if(!(this.bufferOffset < offset && (this.bufferSize + this.bufferOffset) >= (offset + amount))){
				int side = (this.attemptBufferSize - amount) / 2;
				long nbufferOffset = offset - side;
				int nbufferSize = this.attemptBufferSize;
				if(nbufferOffset < 0){
					nbufferOffset = 0;
				}
				if(nbufferOffset + nbufferSize > this.length()){
					nbufferSize = (int)(this.length() - this.bufferOffset);
				}
				buffer(nbufferOffset, nbufferSize);
			}
			this.fileOffset = offset + amount;
			return readBuffedBytes((int)(offset - this.bufferOffset), amount);
		}else{
			this.seek(offset);
			byte[] data = new byte[amount];
			this.read(data, 0, amount);
			return data;
		}
	}
	
	
	public void writeBytes(long offset, byte[] data) throws IOException{
		if(this.fileBuffer != null){
			if(this.bufferOffset <= offset && (this.bufferSize + this.bufferOffset) >= offset){
				int oib = (int)(offset - this.bufferOffset);
				int atsop = (int)((this.bufferSize + this.bufferOffset) - offset);
				int other = (int)((this.bufferSize + this.bufferOffset) - (offset + data.length));
				if(other > -1){
					writeBuffedBytes(oib, data);
				}else{
					byte[] temp = new byte[atsop];
					for(int i = 0; i < atsop; i++){
						temp[i] = data[i];
					}
					writeBuffedBytes(oib, temp);
				}
			}else if(this.bufferOffset > offset && (this.bufferSize + this.bufferOffset)  >= (offset + data.length)){
				/*
				int a = (int)((offset + data.length) - this.bufferOffset);
				int t = data.length - a;
				int p = 0;
				byte[] temp = new byte[a];
				for(int i = t; i < data.length; i++){
					temp[p] = data[t];
					p++;
				}
				writeBuffedBytes(0, temp);
				*/
			}
		}
		super.seek(offset);
		super.write(data, 0, data.length);
		this.fileOffset = offset + data.length;
	}
	
	public void writeBytes(byte[] bytes) throws IOException{
		super.write(bytes, 0, bytes.length);
		this.fileOffset += bytes.length;
	}
	
	public void buffer(long offset, int size) throws IOException{
		this.bufferOffset = offset;
		this.offsetInBuffer = 0;
		this.bufferSize = size;
		this.seek(offset);
		this.fileBuffer = new byte[size];
		super.read(this.fileBuffer, 0, size);
	}
	
	public void killBuffer(){
		this.bufferOffset = -1;
		this.offsetInBuffer = 0;
		this.bufferSize = -1;
		this.fileBuffer = null;
	}
	
	public static class Shift{
		public long timeStamp;
		public long offset;
		public int size;
		public Shift(long offset, int size){
			this.timeStamp = System.currentTimeMillis();
			this.offset = offset;
			this.size = size;
		}
	}
	
	public Offset createOffset(long offset){
		return new Offset(offset);
	}
	
	public Offset createOffset(long offset, long time){
		Offset off = new Offset(offset);
		off.dateStamp = time;
		return off;
	}
	
	public class Offset{
		public long offset;
		public long dateStamp;
		public Offset(long offset){
			this.dateStamp = System.currentTimeMillis();
			this.offset = offset;
		}
		
		public long getOffsetUpdated(){
			long updated = this.offset;
			for(int i = 0; i < shifts.size(); i++){
				Shift shift = shifts.get(i);
				if(shift.timeStamp > this.dateStamp){
					if(updated >= shift.offset){
						updated += shift.size;
					}
				}
			}
			return updated;
		}
	}
	
	public void shift(long offset, long size) throws Exception{
		System.gc();
		long freeMemory = Runtime.getRuntime().freeMemory() - 1000000;
		long oldLength = super.length();
		long newLength = oldLength + size;
		int shiftSize  = (int)(oldLength - offset);
		boolean shortWay = shiftSize < freeMemory;
		if(size < 0){
			shortWay = true;
		}
		if(shortWay){
			System.out.println("short way");
			byte[] data = new byte[shiftSize];
			super.seek(offset);
			super.read(data, 0, shiftSize);
			super.seek(offset + size);
			super.write(data, 0, shiftSize);
			data = null;
		}else{
			System.out.println("long way");
			int amountLeft = shiftSize;
			int defaultChunkSize = (int)freeMemory;
			long fileReadPos = oldLength;
			long fileWritePos = newLength;
			while(amountLeft != 0){
				int chunkSize;
				if(amountLeft > defaultChunkSize){
					chunkSize = defaultChunkSize;
				}else{
					chunkSize = amountLeft;
				}
				fileReadPos -= chunkSize;
				fileWritePos -= chunkSize;
				byte[] temp = new byte[chunkSize];
				super.seek(fileReadPos);
				super.read(temp, 0, chunkSize);
				super.seek(fileWritePos);
				super.write(temp, 0, chunkSize);
				amountLeft -= chunkSize;
			}
		}
		if(size > 0){
			byte[] temp = new byte[(int)size];
			super.seek(offset);
			super.write(temp, 0, temp.length);
		}
		super.setLength(newLength);
		this.shifts.add(new Shift(offset, (int)size));
	}

}
